from enum import Enum

from meross_iot.cloud.abilities import HUB_MTS100_ALL, HUB_MTS100_TEMPERATURE, HUB_MTS100_MODE
from meross_iot.cloud.devices.subdevices.generic import GenericSubDevice
from meross_iot.logger import VALVES_LOGGER as l


class ThermostatMode(Enum):
    COMFORT = 1
    CUSTOM = 0
    ECONOMY = 2
    SCHEDULE = 3


class ThermostatV3Mode(Enum):
    AUTO = 3
    COOL = 2
    CUSTOM = 0
    ECONOMY = 4
    HEAT = 1


class ValveSubDevice(GenericSubDevice):
    mode = None
    #heating = None
    open_window = None
    target_temperature = None
    room_temperature = None
    custom_temperature = None
    comfort_temperature = None
    economy_temperature = None
    max_temperature = None
    min_temperature = None

    type = 'Unknown'

    def __init__(self, subdevice_id, hub, **kwords):
        super().__init__(subdevice_id, hub, **kwords)
        if 'mts100v3' in kwords:
            self.type = 'mts100v3'
        elif 'mts100' in kwords:
            self.type = 'mts100'

    def update_all(self):
        payload = {'all': [{'id': self.id}]}
        res = self._hub.execute_command('GET', HUB_MTS100_ALL, payload)
        my_data = res.get('all')[0]
        self._update_all(my_data)

    def handle_push_event(self, subdevice_data, namespace):
        if namespace == HUB_MTS100_ALL:
            self._update_all(subdevice_data)
        elif namespace == HUB_MTS100_MODE:
            self._update_mode(subdevice_data)
        elif namespace == HUB_MTS100_TEMPERATURE:
            self._update_temperature(subdevice_data)
        else:
            # In any other case, invoke the generic subdevice handler
            super().handle_push_event(subdevice_data, namespace)

    def _update_mode(self, mode):
        mode_state = mode.get('state')
        if self.type == 'mts100':
            self.mode = ThermostatMode(mode_state)
        elif self.type == 'mts100v3':
            self.mode = ThermostatV3Mode(mode_state)

    def _update_temperature(self, temperature_data):
        if 'room' in temperature_data:
            self.room_temperature = temperature_data.get('room')
        if 'currentSet' in temperature_data:
            self.target_temperature = temperature_data.get('currentSet')
        if 'custom' in temperature_data:
            self.custom_temperature = temperature_data.get('custom')
        if 'comfort' in temperature_data:
            self.comfort_temperature = temperature_data.get('comfort')
        if 'economy' in temperature_data:
            self.economy_temperature = temperature_data.get('economy')
        if 'max' in temperature_data:
            self.max_temperature = temperature_data.get('max')
        if 'min' in temperature_data:
            self.min_temperature = temperature_data.get('min')
        if 'openWindow' in temperature_data:
            self.open_window = temperature_data.get('openWindow') == 1

    def _update_all(self, my_data):
        if 'online' in my_data and 'lastActiveTime' in my_data.get('online'):
            self.last_active_time = my_data.get('online').get('lastActiveTime')

        if 'online' in my_data and 'status' in my_data.get('online'):
            self.online = my_data.get('online').get('status') == 1

        if 'togglex' in my_data and 'onoff' in my_data.get('togglex'):
            self.onoff = my_data.get('togglex').get('onoff')

        if 'mode' in my_data:
            self._update_mode(my_data.get('mode'))

        if 'temperature' in my_data:
            temps = my_data.get('temperature')
            self._update_temperature(temps)

            # It's never TRUE: probably is driven by the app comparing the target with the room
            # self.heating = temps.get('heating') == 1

    def __str__(self):
        return "mode={}, room={}, target={}, open window={}".format(self.mode, self.room_temperature,
                                                                    self.target_temperature,
                                                                    self.open_window)



# Common abilities
ALL = 'Appliance.System.All'
ABILITY = 'Appliance.System.Ability'
REPORT = 'Appliance.System.Report'
ONLINE = 'Appliance.System.Online'
WIFI_LIST = 'Appliance.Config.WifiList'
DEBUG = 'Appliance.System.Debug'
TRACE = 'Appliance.Config.Trace'

# Power plug/bulbs abilities
TOGGLE = 'Appliance.Control.Toggle'
TOGGLEX = 'Appliance.Control.ToggleX'
TRIGGER = 'Appliance.Control.Trigger'
TRIGGERX = 'Appliance.Control.TriggerX'
ELECTRICITY = 'Appliance.Control.Electricity'
CONSUMPTIONX = 'Appliance.Control.ConsumptionX'

# Hub
HUB_TOGGLEX = 'Appliance.Hub.ToggleX'
HUB_MTS100_TEMPERATURE = 'Appliance.Hub.Mts100.Temperature'
HUB_MTS100_MODE = 'Appliance.Hub.Mts100.Mode'
HUB_MTS100_ALL = 'Appliance.Hub.Mts100.All'

# Garage opener abilities
GARAGE_DOOR_STATE = 'Appliance.GarageDoor.State'

# Bulbs-only abilities
LIGHT = 'Appliance.Control.Light'



from enum import Enum
from meross_iot.cloud.abilities import *
from meross_iot.cloud.device import AbstractMerossDevice
from meross_iot.logger import BULBS_LOGGER as l
from meross_iot.meross_event import BulbSwitchStateChangeEvent, BulbLightStateChangeEvent


MODE_LUMINANCE = 4
MODE_TEMPERATURE = 2
MODE_RGB = 1
MODE_RGB_LUMINANCE = 5
MODE_TEMPERATURE_LUMINANCE = 6


def to_rgb(rgb):
    if rgb is None:
        return None
    elif isinstance(rgb, int):
        return rgb
    elif isinstance(rgb, tuple):
        red, green, blue = rgb
    elif isinstance(rgb, dict):
        red = rgb['red']
        green = rgb['green']
        blue = rgb['blue']
    else:
        raise Exception("Invalid value for RGB!")

    r = red << 16
    g = green << 8
    b = blue

    return r+g+b


class GenericBulb(AbstractMerossDevice):
    # Bulb state: dictionary of channel-id/bulb-state
    _state = None

    def __init__(self, cloud_client, device_uuid, **kwords):
        self._state = {}
        super(GenericBulb, self).__init__(cloud_client, device_uuid, **kwords)

    def _channel_control_impl(self, channel, status):
        if TOGGLE in self.get_abilities():
            return self._toggle(status)
        elif TOGGLEX in self.get_abilities():
            return self._togglex(channel, status)
        else:
            raise Exception("The current device does not support neither TOGGLE nor TOGGLEX.")

    def _update_state(self, channel, **kwargs):
        with self._state_lock:
            if channel not in self._state:
                self._state[channel] = {}
            for k in kwargs:
                if k == 'onoff':
                    self._state[channel]['onoff'] = kwargs[k]
                elif kwargs[k] is not None:
                    self._state[channel][k] = kwargs[k]

    def _toggle(self, status):
        payload = {"channel": 0, "toggle": {"onoff": status}}
        return self.execute_command("SET", TOGGLE, payload)

    def _togglex(self, channel, status):
        payload = {'togglex': {"onoff": status, "channel": channel}}
        return self.execute_command("SET", TOGGLEX, payload)

    def _handle_push_notification(self, namespace, payload, from_myself=False):
        def fire_bulb_switch_state_change(dev, channel_id, o_state, n_state, f_myself):
            if o_state != n_state:
                evt = BulbSwitchStateChangeEvent(dev=dev, channel_id=channel_id, is_on=n_state,
                                                 generated_by_myself=f_myself)
                self.fire_event(evt)

        def fire_bulb_light_state_change(dev, channel_id, o_state, n_state, f_myself):
            if o_state != n_state:
                evt = BulbLightStateChangeEvent(dev=dev, channel_id=channel_id, light_state=n_state,
                                                generated_by_myself=f_myself)
                self.fire_event(evt)

        with self._state_lock:
            if namespace == TOGGLE:
                on_status=payload['toggle']['onoff'] == 1
                channel_index = 0
                old_state = self._state.get(channel_index)
                new_state = on_status
                self._update_state(channel=0, onoff=on_status)
                fire_bulb_switch_state_change(self, channel_id=0, o_state=old_state, n_state=new_state,
                                              f_myself=from_myself)

            elif namespace == TOGGLEX:
                if isinstance(payload['togglex'], list):
                    for c in payload['togglex']:
                        channel_index = c['channel']
                        on_status = c['onoff'] == 1
                        old_state = self._state.get(channel_index)
                        if old_state is not None:
                            old_state = old_state.get('onoff')
                        self._update_state(channel=channel_index, onoff=on_status)
                        fire_bulb_switch_state_change(self, channel_id=channel_index, o_state=old_state,
                                                      n_state=on_status, f_myself=from_myself)
                elif isinstance(payload['togglex'], dict):
                    channel_index = payload['togglex']['channel']
                    on_status = payload['togglex']['onoff'] == 1
                    old_state = self._state.get(channel_index).get('onoff')
                    if old_state is not None:
                        old_state = old_state.get('onoff')
                    self._update_state(channel=channel_index, onoff=on_status)
                    fire_bulb_switch_state_change(self, channel_id=channel_index, o_state=old_state, n_state=on_status,
                                                  f_myself=from_myself)

            elif namespace == LIGHT:
                channel_index = payload['light']['channel']
                old_state = self._state.get(channel_index)
                new_state = payload['light']
                del new_state['channel']
                self._update_state(channel=channel_index, **new_state)
                fire_bulb_light_state_change(self, channel_id=channel_index, o_state=old_state, n_state=new_state,
                                             f_myself=from_myself)

            elif namespace == REPORT:
                # For now, we simply ignore push notification of these kind.
                # In the future, we might think of handling such notification by caching them
                # and avoid the network round-trip when asking for power consumption (if the latest report is
                # recent enough)
                pass

            else:
                l.error("Unknown/Unsupported namespace/command: %s" % namespace)

    def _get_status_impl(self):
        res = {}
        data = self.get_sys_data()['all']
        if 'digest' in data:
            light_channel = data['digest']['light']['channel']
            res[light_channel] = data['digest']['light']

            for c in data['digest']['togglex']:
                res[c['channel']]['onoff'] = c['onoff'] == 1
        elif 'control' in data:
            res[0]['onoff'] = data['control']['toggle']['onoff'] == 1
        return res

    def _get_channel_id(self, channel):
        # Otherwise, if the passed channel looks like the channel spec, lookup its array indexindex
        if channel in self._channels:
            return self._channels.index(channel)

        # if a channel name is given, lookup the channel id from the name
        if isinstance(channel, str):
            for i, c in enumerate(self.get_channels()):
                if c['devName'] == channel:
                    return c['channel']

        # If an integer is given assume that is the channel ID
        elif isinstance(channel, int):
            return channel

        # In other cases return an error
        raise Exception("Invalid channel specified.")

    def get_status(self, channel=0):
        # In order to optimize the network traffic, we don't call the get_status() api at every request.
        # On the contrary, we call it the first time. Then, the rest of the API will silently listen
        # for state changes and will automatically update the self._state structure listening for
        # messages of the device.
        # Such approach, however, has a side effect. If we call TOGGLE/TOGGLEX and immediately after we call
        # get_status(), the reported status will be still the old one. This is a race condition because the
        # "status" RESPONSE will be delivered some time after the TOGGLE REQUEST. It's not a big issue for now,
        # and synchronizing the two things would be inefficient and probably not very useful.
        # Just remember to wait some time before testing the status of the item after a toggle.
        c = self._get_channel_id(channel)
        if self._state == {}:
            current_state = self._get_status_impl()
            with self._state_lock:
                self._state = current_state
        return self._state[c]

    def get_channels(self):
        return self._channels

    def get_channel_status(self, channel):
        ch_id = self._get_channel_id(channel)
        c = self._get_channel_id(ch_id)
        return self.get_status(c)

    def turn_on_channel(self, channel):
        ch_id = self._get_channel_id(channel)
        c = self._get_channel_id(ch_id)
        return self._channel_control_impl(c, 1)

    def turn_off_channel(self, channel):
        ch_id = self._get_channel_id(channel)
        c = self._get_channel_id(ch_id)
        return self._channel_control_impl(c, 0)

    def turn_on(self, channel=0):
        ch_id = self._get_channel_id(channel)
        c = self._get_channel_id(ch_id)
        return self._channel_control_impl(c, 1)

    def turn_off(self, channel=0):
        ch_id = self._get_channel_id(channel)
        c = self._get_channel_id(ch_id)
        return self._channel_control_impl(c, 0)

    def set_light_color(self, channel=0, rgb=None, luminance=-1, temperature=-1, capacity=None):
        """
        :param channel: Channel to control (for bulbs it's usually 0)
        :param rgb: (red,green,blue) tuple, where each color is an integer from 0-to-255
        :param luminance: Light intensity (at least on MSL120). Varies from 0 to 100
        :param temperature: Light temperature. Can be used when rgb is not specified.
        :param capacity: Old parameter. Kept for compatibility reasons. Will be removed in the future.
        :return:
        """
        ch_id = self._get_channel_id(channel)

        if rgb is not None and temperature != -1:
            l.error("You are trying to set both RGB and luminance values for this bulb. It won't work!")

        # Prepare a basic payload
        payload = {
            'light': {
                'channel': ch_id,
                'gradual': 0
            }
        }

        mode = 0
        if self.supports_mode(MODE_RGB) and rgb is not None:
            # Convert the RGB to integer
            color = to_rgb(rgb)
            payload['light']['rgb'] = color
            mode = mode | MODE_RGB

        if self.supports_mode(MODE_LUMINANCE) and luminance != -1:
            payload['light']['luminance'] = luminance
            mode = mode | MODE_LUMINANCE

        if self.supports_mode(MODE_TEMPERATURE) and temperature != -1:
            payload['light']['temperature'] = temperature
            mode = mode | MODE_TEMPERATURE

        payload['light']['capacity'] = mode

        self.execute_command(command='SET', namespace=LIGHT, payload=payload)

    def supports_mode(self, mode):
        return (self.get_abilities().get(LIGHT).get('capacity') & mode) == mode

    def get_light_color(self, channel=0):
        ch_id = self._get_channel_id(channel)
        return self.get_status(channel=ch_id)

    def get_power_consumption(self):
        return None

    def get_electricity(self):
        return None

    def is_rgb(self):
        return self.supports_mode(MODE_RGB)

    def is_light_temperature(self):
        return self.supports_mode(MODE_TEMPERATURE)

    def supports_luminance(self):
        return self.supports_mode(MODE_LUMINANCE)

    def __str__(self):
        base_str = super().__str__()
        with self._state_lock:
            if not self.online:
                return base_str
            channels = "Channels: "
            channels += ",".join(["%d = %s" % (k, "ON" if v else "OFF") for k, v in enumerate(self._state)])
            return base_str + "\n" + "\n" + channels



from abc import ABC, abstractmethod
from threading import RLock

from meross_iot.cloud.exceptions.OfflineDeviceException import OfflineDeviceException
from meross_iot.cloud.timeouts import LONG_TIMEOUT, SHORT_TIMEOUT
from meross_iot.cloud.abilities import *
from meross_iot.logger import DEVICE_LOGGER as l
from meross_iot.meross_event import DeviceOnlineStatusEvent


class AbstractMerossDevice(ABC):
    # Device status + lock to protect concurrent access
    _state_lock = None
    online = False

    # Device info and connection parameters
    uuid = None
    app_id = None
    name = None
    type = None
    hwversion = None
    fwversion = None

    # Cached list of abilities
    _abilities = None

    # Cloud client: the object that handles mqtt communication with the Meross Cloud
    __cloud_client = None

    # Data structure for firing events.
    __event_handlers_lock = None
    __event_handlers = None

    def __init__(self, cloud_client, device_uuid, **kwargs):
        self.__cloud_client = cloud_client
        self._state_lock = RLock()
        self.__event_handlers_lock = RLock()
        self.__event_handlers = []

        self.uuid = device_uuid

        if "channels" in kwargs:
            self._channels = kwargs['channels']

        # Information about device
        if "devName" in kwargs:
            self.name = kwargs['devName']
        if "deviceType" in kwargs:
            self.type = kwargs['deviceType']
        if "fmwareVersion" in kwargs:
            self.fwversion = kwargs['fmwareVersion']
        if "hdwareVersion" in kwargs:
            self.hwversion = kwargs['hdwareVersion']
        if "onlineStatus" in kwargs:
            self.online = kwargs['onlineStatus'] == 1

    def handle_push_notification(self, namespace, payload, from_myself=False):
        # Handle the ONLINE push notification
        # Leave the rest to the specific implementation
        if namespace == ONLINE:
            old_online_status = self.online
            status = payload['online']['status']
            if status == 2:
                with self._state_lock:
                    self.online = False
            elif status == 1:
                with self._state_lock:
                    self.online = True
            else:
                l.error("Unknown online status has been reported from the device: %d" % status)

            # If the online status changed, fire the corresponding event
            if old_online_status != self.online:
                evt = DeviceOnlineStatusEvent(self, self.online)
                self.fire_event(evt)
        else:
            self._handle_push_notification(namespace, payload, from_myself=from_myself)

    def register_event_callback(self, callback):
        with self.__event_handlers_lock:
            if callback not in self.__event_handlers:
                self.__event_handlers.append(callback)
            else:
                l.debug("The callback you tried to register is already present.")
                pass

    def unregister_event_callback(self, callback):
        with self.__event_handlers_lock:
            if callback in self.__event_handlers:
                self.__event_handlers.remove(callback)
            else:
                l.debug("The callback you tried to unregister is not present.")
                pass

    def fire_event(self, eventobj):
        for c in self.__event_handlers:
            try:
                c(eventobj)
            except:
                l.exception("Unhandled error occurred while executing the registered event-callback")

    @abstractmethod
    def _handle_push_notification(self, namespace, payload, from_myself=False):
        """
        Handles push messages for this device. This method should be implemented by the base class in order
        to catch status changes issued by other clients (i.e. the Meross app on the user's device).
        :param namespace:
        :param message:
        :param from_myself: boolean flag. When true, it means that the notification is generated in response to a
        command that was issued by this client. When false, it means that another client generated the event.
        :return:
        """
        pass

    @abstractmethod
    def get_status(self):
        pass

    def execute_command(self, command, namespace, payload, callback=None, timeout=SHORT_TIMEOUT):
        with self._state_lock:
            # If the device is not online, what's the point of issuing the command?
            if not self.online:
                raise OfflineDeviceException("The device %s (%s) is offline. The command cannot be executed" %
                                             (self.name, self.uuid))

        return self.__cloud_client.execute_cmd(self.uuid, command, namespace, payload, callback=callback, timeout=timeout)

    def get_sys_data(self):
        return self.execute_command("GET", ALL, {})

    def get_abilities(self):
        # TODO: Make this cached value expire after a bit...
        if self._abilities is None:
            self._abilities = self.execute_command("GET", ABILITY, {})['ability']
        return self._abilities

    def get_report(self):
        return self.execute_command("GET", REPORT, {})

    def get_wifi_list(self):
        if WIFI_LIST in self.get_abilities():
            return self.execute_command("GET", WIFI_LIST, {}, timeout=LONG_TIMEOUT)
        else:
            l.error("This device does not support the WIFI_LIST ability")
            return None

    def get_trace(self):
        if TRACE in self.get_abilities():
            return self.execute_command("GET", TRACE, {})
        else:
            l.error("This device does not support the TRACE ability")
            return None

    def get_debug(self):
        if DEBUG in self.get_abilities():
            return self.execute_command("GET", DEBUG, {})
        else:
            l.error("This device does not support the DEBUG ability")
            return None

    def supports_consumption_reading(self):
        return CONSUMPTIONX in self.get_abilities()

    def supports_electricity_reading(self):
        return ELECTRICITY in self.get_abilities()

    def supports_light_control(self):
        return LIGHT in self.get_abilities()

    def __str__(self):
        basic_info = "%s (%s, HW %s, FW %s): " % (
            self.name,
            self.type,
            self.hwversion,
            self.fwversion
        )

        return basic_info
